//****************************************************************************
//                   Sprite handling module, animation                        
//****************************************************************************
//  Written by : Vojvo & Gunny
//  Language   : VisuC6 + DX6.1
//  Version    : 1.1.2
//  Last update: 08.11.2004     
//****************************************************************************
/*
// changes:
   1.1.2 / opraveny vpisy o bodovch kouzel
   1.1.1  / u hrace zavedeny promenne lastseenX (pro Murrayho)
	         pokus o opravu havarie pri plivani za chuze do zdi 
// 1.0.1  - snaha o prvni plnou verzi
// 0.99.6 - uprava skoku na dvoji gravitaci
// 0.99.5 - odkomentovan repelent=0 v Ressurection
//        - prohozeni bodovych a powerup bonusu
//        - ukazatele trvani casove omezenych kouzel 
//          (nesmrtelnost, neviditelnost, krehke bubliny, mrazik)
//        - zobrazovani zmenenych statistik jinou barvou nez nezmenenych
// 0.99.4 - repelent
// 0.11.1 - Player se pri zmene smeru za chuze okamzite zastavi a obrati
//        - Player muze plivnout i kdyz je mimo okraje - clipped
// 0.10.2 / pripraveno blikani bufferu
//        - oprava situace, kdy pri plivnuti, kdyz player dobrzdoval, klouzal
//        - kouzla krehke bubliny, vymena bodu, pozic a vlastnosti
//        - drak po doskoku pri aktivni chuzi nezabrzdi (pokud nefiruje)
// 0.10.1 - pridana dalsi kouzla (mrazik, +time)
//        - rozsiren vypis statistik
//        - zvuky stereo
//        - Player::CheckState predelan na odecitani ruzneho poctu stepu
// 0.9.3 - upraven test na posouvani sprita pri non-stable dopadu 
//         na roh plosinky
//       - zprovoznena dalsi kouzla
// 0.9.2 - funguje stav playera RESURRECTED
//       - funguje stav playera INVISIBLE
// 0.9.1 - funguji nektera kouzla
// 0.8.2 - pridany zvuky
// 0.8.1 - spravna barva kapek u hrace
*/

#include <stdio.h>
#include <string.h>
#include "consts.h"
#include "exceptions.h"
#include "graphdd.h"
#include "surfacemanager.h"
#include "keyboard.h"
#include "map.h"
#include "game.h"
#include "font.h"
#include "message.h"
#include "wsound.h"
#include "gonio.h"
#include "property.h"
#include "framesv.h"
#include "sprites.h"


#define AT_BOTTOM      0    // navr.hod. IsStable, kdyz je to na dne

#define STATBAR_WIDTH         30
#define STATBAR_HEIGHT        2
#define STATBAR_TOTALHEIGHT   15
#define STATBAR_BAR1Y         1
#define STATBAR_BAR2Y         4
#define STATBAR_BAR3Y         7
#define STATBAR_BAR4Y         10


//#########################################################################
//--------------------------------- Sprite --------------------------------
//#########################################################################
Sprite::Sprite (int onx, int ony, Property *prop)
{
   type = COMMON;
   gfxmask = prop->tintmask;
   id = prop->id;
   fTgs = prop->fTgs;
   direction = DIR_L;
   celw = prop->r16h->dwCelWidth;
   celh = prop->r16h->dwCelHeight;
   celr = prop->r16h->byCelRow;
   xsize = prop->xsize;
   ysize = prop->ysize;
   b_miny = PLAYFIELD_MIN_Y - ysize;
   b_maxy = PLAYFIELD_MAX_Y - ysize;
   b_maxx = PLAYFIELD_MAX_X - xsize;
   eb_maxy = PLAYFIELD_EMAX_Y - ysize;
   xoff = prop->xoff;
   yoff = prop->yoff;
   frames.frames = &(prop->frames);
   frames.setDirection(DIR_L);
   frames.setChain(CHAIN_STAND);
   jumpNext();   
   gfx = prop->surf;
   bubgfx = prop->bubsurf;
   collision = 0;
   gravity = DFT_GRAVITY;
   newx = oldx = onx + xoff;
   newy = oldy = ony + yoff;
   mul_x = newx << SHIFT_MOVE + MUL_HALF;
   mul_y = newy << SHIFT_MOVE + MUL_HALF;
   movex = movey = 0;
   stable_l = prop->stablel;
   stable_r = prop->stabler;
   hitx = prop->hitx;
   hity = prop->hity;
   topx = prop->speed;
   topy = MAX_FALLSPEED;
   bottomx = -topx;
   bottomy = -JUMP_STRENGTH;
   sort_priority = 0;
   is_stable = 0;
   clipped = false;
   frameLeft = 0;
   frameTop = 0;
}

//-------------------------------------------------------------------------
// animace
//-------------------------------------------------------------------------
void Sprite::jumpNext()
{
   frames.jumpNext();
   if (frames.chainid == CHAIN_BUBBLE || frames.chainid == CHAIN_SCRAPE)
   {
      frameLeft = BUB_XSIZE * frames.frameid;
      frameTop = 0;
   }
   else
   {
      frameLeft = (frames.frameid % celr) * celw;
      frameTop = (frames.frameid / celr) * celh;
      frameRect.left = frameLeft;
      frameRect.top = frameTop;
      frameRect.right = frameLeft + xsize;  
      frameRect.bottom = frameTop + ysize;  
   }
}

//=========================================================================
void Sprite::Paint(Surface dest)
{
   int y = newy;
   int ybottom = newy + ysize - 1, xright = newx + xsize - 1;
   if (newy < PLAYFIELD_MIN_Y)
   {
      frameRect.top = frameRect.bottom - (newy - b_miny);
      y = PLAYFIELD_MIN_Y;
      clipped = true;
   }
   else if (newy >= b_maxy)
   {
      clipped = true;
      frameRect.bottom = frameRect.top + PLAYFIELD_MAX_Y-newy;
      ybottom = PLAYFIELD_MAX_Y-1;
   }
   else if (clipped)
   {
      frameRect.left = frameLeft;
      frameRect.top = frameTop;
      frameRect.right = frameLeft + xsize;  
      frameRect.bottom = frameTop + ysize;  
      clipped = false;
   }
   
   if (!clipped || frameRect.bottom > frameRect.top)
      putcel(newx,y,dest,frameRect,gfx,DDBLTFAST_SRCCOLORKEY);
   
   SET_USEDMAP(newx,y);
   SET_USEDMAP(xright,y);
   SET_USEDMAP(newx,ybottom);
   SET_USEDMAP(xright,ybottom);
   oldx = newx;
   oldy = newy;
}

//========================================================================
int Sprite::Move()
{
   int result = 0;
   
   mul_x += movex;
   newx = mul_x >> SHIFT_MOVE;
   if (newx<PLAYFIELD_MIN_X) {
      newx = PLAYFIELD_MIN_X;
      mul_x = newx << SHIFT_MOVE + MUL_HALF;
      movex = 0;
      result |= SP_COL_L;
   }
   else if (newx>b_maxx) {
      newx = b_maxx;
      mul_x = newx << SHIFT_MOVE + MUL_HALF;
      movex = 0;
      result |= SP_COL_R;
   }
   mul_y += movey;
   newy = mul_y >> SHIFT_MOVE;
   if (newy<PLAYFIELD_EMIN_Y) {
      newy = PLAYFIELD_EMIN_Y;
      mul_y = newy << SHIFT_MOVE;
      movey = 0;
      result |= SP_COL_U;
   }
   else if (newy>eb_maxy) {
      frameRect.bottom = frameRect.top + ysize;
      newy = PLAYFIELD_EMIN_Y;
      mul_y = newy << SHIFT_MOVE;
      //    movey = 0;
      result |= SP_COL_D;
   }
   collision |= result;
   return result;
}
//==========================================================================
void Sprite::Clear(Surface dest, Surface src)
{
}

//==========================================================================
void Sprite::SlowIt(int by_x, int by_y)
{
   if (movex > 0)
   {
      if (movex > by_x)
         movex -= by_x;
      else
         movex = 0;
   }
   else
   {
      if (movex < (-by_x))
         movex += by_x;
      else
         movex = 0;
   }
   if (movey > 0)
   {
      if (movey > by_y)
         movey -= by_y;
      else
         movey = 0;
   }
   else
   {
      if (movey < (-by_y))
         movey += by_y;
      else
         movey = 0;
   }
}

//==========================================================================
int Sprite::Slow  (int by_x, int by_y)
{
   int result = 0;
   
   movex -= by_x;
   movey -= by_y;
   if (movex < bottomx) {
      movex = bottomx;
      result = 1;
   }
   if (movey < bottomy) {
      movey = bottomy;
      result = 1;
   }
   return result;
}
//==========================================================================
int Sprite::Haste  (int by_x, int by_y)
{
   int result = 0;
   
   movex += by_x;
   movey += by_y;
   if (movex > topx) {
      movex = topx;
      result = 1;
   }
   if (movey > topy) {
      movey = topy;
      result = 1;
   }
   return result;
}

//==========================================================================
int Sprite::IsStable()
{
   int result = 0;
   int parx,pary,top_move,y;
   
   // je stabilni ?
   if (newy >= PLAYFIELD_MAX_Y)
   {
      result = AT_BOTTOM;
   }
   else
   {
      y = newy;
      // podle smeru, x-bod, ktery musi byt stabilni
      if (direction == DIR_R)
         parx = newx+stable_r;
      else
         parx = newx+stable_l;
      
      pary = y+ysize;
      top_move = (topy >> SHIFT_MOVE);
      //
      result=map_value(parx,pary,map);
      if (abs(movey)<MUL_MOVE)
      {
         result = result && !map_value(parx,pary-2,map);
         if (result && map_value(parx,pary-1,map))
         {
            newy -= 1;
            mul_y -= MUL_MOVE;
         }
      }
      else
      {
         result = result && !map_value(parx,pary-top_move-2,map);
         if (result && map_value(parx,pary-top_move-1,map))
         {
            newy -= 1;
            mul_y -= MUL_MOVE;
         }
      }
   }
   is_stable = result;
   return result;
}
//==========================================================================
int Sprite::CantGoLR()
{
   int result,result3,parx1,parx2,pary,y;
   
   result3 = 0;
   y = newy; 
   if (is_stable)
   {
      pary = y+ysize-(topy>>SHIFT_MOVE)-1;
   }
   else
   {
      pary = y+ysize-1;
   }
   
   if (direction == DIR_R)
   {
      parx1 = newx+xsize-1;
      parx2 = newx+stable_r;
   }
   else
   {
      parx1 = newx;
      parx2 = newx+stable_l;
   }
   
   if (map_value(parx1,pary,map) != 0)        // nohy
   {
      if (map_value(parx2,pary,map) != 0)
      {
         result3 = 0;
      }
      else
      {
         if (map_value(parx1,pary+MAP_CELY,map) == 0)
            if (movey > 0)
               result3 = -1; // v souctu vynuluje nutne nasledujici result4 
            else
            {
               if (direction == DIR_L)
               {
                  if (map_value(parx1+MAP_CELX,pary+MAP_CELY,map) == 0)
                     result3 = -1;
                  else
                     result3 = 1;
               }
               else
               {
                  if (map_value(parx1-MAP_CELX,pary+MAP_CELY,map) == 0)
                     result3 = -1;
                  else
                     result3 = 1;
               }
            }
         else
            result3 = 1;
      }
   }
   
   // --- novy Roryho test ---
   int result4 = 0;
   int under1 = under(parx1,pary,map,8);
   int under2 = under(parx2,pary,map,under1);
   if (under1 > under2)
   {
      result4 = 1;
   }
   
#ifdef FILEDEBUG
   if (type == MONSTER)
      fprintf(fw,"%p -> (%d,%d):%d R:%d TEST:%d,%d UNDER %d,%d\n",this,parx1,
      parx2,newy+ysize-pary,(direction > 0),result3,result4,under1,under2);
#endif
   
   result = result3 + result4;
   if (result)
   {
      if (direction == DIR_R)
         collision |= SP_COL_R;
      else
         collision |= SP_COL_L;
   }
   return (result);
}



//##########################################################################
//------------------------------- Creature ---------------------------------
//##########################################################################

//===========================================================================
void Creature::MapReflex()
{
   IsStable();
   
   // osetreni "prochazeni zdi" doleva a doprava
   if (CantGoLR())
   {
      // kdyz jde doleva je treba ho posunout o jeden zpatky
      if (direction == DIR_L)
		{
         newx = (newx/MAP_CELX+1) * MAP_CELX;
		}
      else
		{
         newx = (newx/MAP_CELX+1) * MAP_CELX - (xsize % MAP_CELX);
		}
      mul_x = (newx << SHIFT_MOVE) + MUL_HALF;
      movex = 0;
   }
   
   // osetreni dopadu na plosinku
   if ((action & SPMA_JUMP) && movey > 0 && is_stable)
   {
      sound_device->Play(2,newx+xsize/2);
      newy = (newy/MAP_CELY) * MAP_CELY + yoff;
      mul_y = newy << SHIFT_MOVE;
      movey = 0;
      action &= ~(SPMA_LEFT + SPMA_RIGHT);
      if (!(action & SPMA_FIRE))
      {
         frames.setChain(CHAIN_STAND);
//         action &= ~(SPMA_LEFT + SPMA_RIGHT);
      }
      else
      {
         movex = 0;
         old_chain = CHAIN_STAND;         
      }
      action ^= SPMA_JUMP;
   }
   
   // osetreni spadnuti na konci plosinky
   if (!(action & SPMA_JUMP) && !is_stable)
   {
      action |= SPMA_JUMP;
      old_chain = frames.chainid;
      if (!(action & SPMA_FIRE)) frames.setChain(CHAIN_FALL);
   }
}

//===========================================================================
void Creature::Reincarnate()
{
   movex = 0;
   movey = 0;
   action = 0;
}


//##########################################################################
//-------------------------------- Player ----------------------------------
//##########################################################################
Player::Player(Property *prop, Font *font) : Creature (0,0,prop)
{
   type = PLAYER;
   dik_left = dik_right = dik_jump = dik_fire = 0;
   sc_last = 0;
   sheet.font = font;
   litetint = prop->tint;
   hardtint = prop->tint2;
   state = 0;
   time_resurrected.options(15000,0,30,true,false);
   time_invisible.options(15000,0,30,true,false); 
   time_selfblowing.options(15000,0,30,true,false);
   time_resurrected.decrease(time_resurrected.timecurrent);
   time_invisible.decrease(time_invisible.timecurrent);
   time_selfblowing.decrease(time_selfblowing.timecurrent);
   updateable = 1;
   //
   jump_strength = JUMP_STRENGTH;
   bottomy = -MOVE600;
   acceleration = MOVE_ACCEL;
   bubble_speed = SPEED_BUBBLE;
   //bubble_delay = 500;
   repelent = 0;
   gravityLow = gravity;
   gravityHigh = gravity*2;
}

//-------------------------------------------------------------------------
// usadi hrace na spravne misto do levelu a vubec ho pripravi na hru V 2004
//-------------------------------------------------------------------------
void Player::Place(int x, int y)
{
	// takze pozice v pixelech
   newx = x * MAP_CELX;
   newy = y * MAP_CELY + PLAYFIELD_TOP;

   frames.setChain(CHAIN_STAND);
	// ktery smerem se diva
   if (x>15) 
   {
      direction=DIR_L;
      frames.setDirection(DIR_L);
		// drak nema obvyklych 40pix, tak at je to presne
		newx += 40 - xsize;
   }
   else 
   {
      direction=DIR_R;
      frames.setDirection(DIR_R);
   }
   jumpNext();
   
   oldx = newx;
   mul_x = newx << SHIFT_MOVE;
   oldy = newy;
   mul_y = newy << SHIFT_MOVE;
   sheet.energy = sheet.maxenergy;
   Reincarnate();
   state &= IMMORTAL;

	// pro jistotu nejak nastavim lastseen
	// (nemelo by to mit vliv, protoze to je jen na prvni pruchod smyckou)
	lastseenx = newx;
	lastseeny = newy;

}


//-------------------------------------------------------------------------
// nastavuje klavesy
//-------------------------------------------------------------------------
void Player::SetKeys(BYTE dikl,BYTE dikr,BYTE dikj,
                     BYTE dikf)
{
   dik_left = dikl;
   dik_right = dikr;
   dik_jump = dikj;
   dik_fire = dikf;
}

//-------------------------------------------------------------------------
// reaguje na aktualni stav playera
//-------------------------------------------------------------------------
void Player::stateReflex()
{
   if (!(state & INBUBBLE))
   {
      // kdyz skace, zpomaluje ho gravitace
      if (action & SPMA_JUMP)
      {
         if (movey > 0 && frames.chainid == CHAIN_JUMP)
         {
            if (action & SPMA_FIRE)
            {
               old_chain = CHAIN_FALL;
            }
            else
            {
               frames.setChain(CHAIN_FALL);
               jumpNext();
            }
         }
         Haste(0,gravity);
      }
      // ukonceni firingu; je treba nastavit ukazovatko na puvodni framy
      if ((action & SPMA_FIRE) && frames.chainid != CHAIN_FIRE)
      {
         frames.setChain(old_chain);
         action &= ~(SPMA_FIRE | SPMA_LEFT | SPMA_RIGHT);
      }
      //kdyz probiha fire neprijma se zadna akce
      if (!(action & SPMA_FIRE))
      {
         if (action & SPMA_LEFT)
         {
            if (direction == DIR_R && !(action & SPMA_JUMP))
               movex = 0;
            Slow(acceleration,0);
            if (movex < 0 && direction == DIR_R)  // otoceni spritu
            {
               direction = DIR_L;
               frames.setDirection(DIR_L);
               jumpNext();
               newx += stable_r-stable_l;
               mul_x = newx << SHIFT_MOVE + MUL_HALF;
            }
         }
         else if (action & SPMA_RIGHT)
         {
            if (direction == DIR_L && !(action & SPMA_JUMP))
               movex = 0;
            Haste (acceleration,0);
            if (movex > 0 && direction == DIR_L)  // otoceni spritu
            {
               direction = DIR_R;
               frames.setDirection(DIR_R);
               jumpNext();
               newx -= stable_r-stable_l-1;
               mul_x = newx << SHIFT_MOVE + MUL_HALF;
            }
         }
      }
   }
}

//==========================================================================
void Player::KeyReflex()
// reakce na stisk klaves:  sc_left .... doleva
//                          sc_right ... doprava
//                          sc_jump .... skok
//                          sc_fire .... vypusteni bubliny     
{
   BYTE last = 0;
   Player *otherplayer;
   
   // kdyz hrac pusti klavesu fire povol jeji opetovny stisk
   if (fk_pressed && !KEYDOWN(kbd.diKeys,dik_fire))
      fk_pressed = false;
   
   if (action & SPMA_JUMP && gravity == gravityLow && !KEYDOWN(kbd.diKeys,dik_jump))
   {
      gravity = gravityHigh;
   }
   if (!(state & INBUBBLE))
   {
		// neni v bubline
      //kdyz probiha fire neprijma se zadna akce
      if (!(action & SPMA_FIRE))
      {
         // hrac drzi klavesu "vlevo"
         if (KEYDOWN(kbd.diKeys,dik_left)) 
         {
            if (action==0) 
            {
               frames.setChain(CHAIN_WALK);
               jumpNext();
            }
            action |= SPMA_LEFT;
            last |= SCL_LEFT;
         }
         else
         {
            action &= ~SPMA_LEFT;
         }
         // hrac drzi klavesu "vpravo"
         if (KEYDOWN(kbd.diKeys,dik_right)) 
         {
            if (action==0) 
            {
               frames.setChain(CHAIN_WALK);
               jumpNext();
            }
            action |= SPMA_RIGHT;
            last |= SCL_RIGHT;
         }
         else
         {
            action &= ~SPMA_RIGHT;
         }
         // hrac stiskl klavesu "skok" a pajdulak prave neskace
         if (KEYDOWN(kbd.diKeys,dik_jump) && !(action & SPMA_JUMP))
         {
            sound_device->Play(3,newx+xsize/2);
            action |= SPMA_JUMP;
            frames.setChain(CHAIN_JUMP);
            jumpNext();
            Slow (0,jump_strength);
            gravity = gravityLow;
            last |= SCL_JUMP;
         }
         // hrac stiskl klavesu "fire" a nedrzi ji uz od minuleho fire
         // a neni orezan - mimo hraci pole
			// to by mohl neco vyplivnout =)
         if (KEYDOWN(kbd.diKeys,dik_fire) && !fk_pressed)
         {
				// fajn... tak kam s bublinou?
            long x = newx;
            long y = newy;
				// vhodna pozice X
            if (direction == DIR_L)
               x = x + xsize - BUB_XSIZE;
				// V. 2004 - kontrola na kraji obrazovky
				if (x > (PLAYFIELD_MAX_X - BUB_XSIZE))
				{
					x = PLAYFIELD_MAX_X - BUB_XSIZE;
				}
				if (x < (PLAYFIELD_MIN_X))
				{
					x = PLAYFIELD_MIN_X;
				}
				// eof 2003

				// vhodna pozice Y
            if (newy < PLAYFIELD_MIN_Y && newy > 10)
            {
               y = 40;
            }
            else if (newy > (PLAYFIELD_MAX_Y - ysize) && newy < 430)
            {
               y = 400;
            }
            else if (clipped)
               y = 0;
            if (y > 0)
            {
               action |= SPMA_FIRE;
               old_chain = frames.chainid;
               if (old_chain == CHAIN_WALK)
                  old_chain = CHAIN_STAND;
               frames.setChain(CHAIN_FIRE);
               if (!(action & SPMA_JUMP)) 
               {
                  // zrusime chuzi doleva|doprava
                  movex=0;
               }
               fk_pressed = true;
					// ma jeste co plivat?
               if (!sheet.energy)
               {
                  sound_device->Play(29,newx+xsize/2); //skytaci sampl
                  drops->OneDrop(newx+((direction)?34:6),y+8,0,-MOVE50,dropcolor,0,10);
               }
               else
               { 
						// jo, rezervni bubliny by byly
                  int bubid = ID_BUBBLE;
                  if (s_solidbubbles)
                     if (this->id == ID_PLAYER1)
                        bubid = ID_SOLIDBUBBLEG;
                     else
                        bubid = ID_SOLIDBUBBLEB;
                  Bubble *pom = new Bubble(x,y,properties.member[bubid]);
                  pom->tint = litetint;
                  pom->frames.setChain(CHAIN_RELEASE);
                  pom->jumpNext();
                  // uprava na bonus 'krehke bubliny'
                  if (time_selfblowing.timecurrent > 0)
                  {
                     pom->selfblowing = 1;
                     pom->tint = hardtint;
                  }               
                  sound_device->Play(0,newx+xsize/2);
                  pom->whose = this;
                  pom->dropcolor = dropcolor;
                  sheet.energy--;
                  sheet.energy_changed = 2;
                  if (ARRAY_FULL(qbubbles_ctrl))
                  {
                     delete pom;
                  }
                  else
                  {
                     ARRAY_ADD(qbubbles,qbubbles_ctrl,pom)
                        if (direction == DIR_R)
                           pom->movex = bubble_speed;
                        else
                           pom->movex = -bubble_speed;
                  }
               }            
               last |= SCL_FIRE;
            }
         }
      }
      // hrac nedrzi zadnou klavesu
      if (!last)
      {
         // kdyz neni skok nebo fire tak zrus chuzi
         if (!(action & SPMA_JUMP))
         {
            if (!(action & SPMA_FIRE)) 
            {
               movex = 0;
               action = 0;
               if (sc_last)
                  frames.setChain(CHAIN_STAND);
            }
            else
            {
               action &= ~(SPMA_LEFT + SPMA_RIGHT);
               old_chain = CHAIN_STAND;
            }
         }
         else
            action &= ~(SPMA_LEFT + SPMA_RIGHT);
      }
      else
      {
         // zhmotneni
         if (state & NOTTAKEN)
         {
            sound_device->Play(12,newx+xsize/2);
            state ^= UNCONTROLLED;
            // nastaveni nezranitelnosti
            state |= RESURRECTED;
            time_resurrected.increase(1500);
            otherplayer = (this == player1) ? player2 : player1;
            --player1->sheet.lives;
            lives_chg = 2;
            if ((otherplayer->state & NOTTAKEN) && !player1->sheet.lives)
            {
               // Don't let reincarnate otherplayer 
               otherplayer->state |= DEAD;
               ARRAY_REMOVE(monsters,monsters_ctrl,otherplayer);
            }
         }
      }
   }
   else
   {
		// je v bubline
      if (KEYDOWN(kbd.diKeys,dik_fire) && !fk_pressed)
      {
         state |= WANNALEAVE;
         fk_pressed = true;
			// vrzanim prozradi Murraymu svoji pozici
			lastseenx = newx;
			lastseeny = newy;
      }
   }
   sc_last = last;
}

//==========================================================================
int Player::CollideWith(Sprite *what)
{
   int result, value;
   int diff_x,diff_y;
   
   result = 0;
   diff_x = what->newx - newx;
   diff_y = what->newy - newy;
   switch (what->type)
   {
   case BONUS: if (diff_x < 0 && -diff_x < (what->xsize-PLAYERBONUS_HITTOLX) || 
                   diff_x >= 0 && diff_x < (xsize-PLAYERBONUS_HITTOLX))
                  if (diff_y < 0 && -diff_y < (what->ysize-PLAYERBONUS_HITTOLY) || 
                      diff_y >= 0 && diff_y < (ysize-PLAYERBONUS_HITTOLY))
                  {
                     result = 1;
                     value = ((Bonus *)what)->value;
                     sheet.AddPoints(value,what->newx,what->newy-16);
                     ((Bonus *)what)->counter = BONUS_LIFE;
                  }
                  break;
   case MONSTER: if (diff_x < 0 && -diff_x < (what->xsize-what->hitx) || 
                     diff_x >= 0 && diff_x < (xsize-what->hitx))
                    if (diff_y < 0 && -diff_y < (what->ysize-what->hity) || 
                        diff_y >= 0 && diff_y < (ysize-PLAYERMONSTER_HITTOLY))
                    {
                       result = 1;
                       Die();
                    }
                    break;
   }
   return result;
}


//==========================================================================

void Player::Die() 
{
   state &= ~(INVISIBLE);
   frames.setChain(CHAIN_DYING);
   sound_device->Play(19,newx+xsize/2);
   state = state | DEAD;
   movex = 0;
   movey = 0;
   sheet.koef = 1;
   UFO *ufo;
   for (int i=0; i<UFOcnt; i++)
   {
      ufo = UFOs[i];
      if (ufo->id==ID_MURRAYB || ufo->id==ID_MURRAYG)
         if (((Murray *)ufo)->target==this)
            ufo->Melt();
   }
}

//==========================================================================
void Player::Resurrection()
{
   state |= UNCONTROLLED;
   state &= ~(INACTIVE);
   
   time_resurrected.set(0);
   time_invisible.set(0);
   time_selfblowing.set(0);
   repelent = 0; //zatim ho necham jako jednu z vlastnosti
   
   frames.setChain(CHAIN_STAND);
   Reincarnate();
}

//==========================================================================
void Player::Paint(Surface dest)
{
   RECT r;
   int y = newy;
   int ybottom = newy + ysize - 1, xright = newx + xsize - 1;
	// je normalne na plose
   if (newy >= PLAYFIELD_MIN_Y + STATBAR_TOTALHEIGHT && !(state & NOTTAKEN))
   {
      r.top = STATBAR_BAR4Y;
      r.bottom = STATBAR_BAR4Y + STATBAR_HEIGHT; 
      r.left = 0; 
      r.right = frozen.xcurrent;
      if (r.left < r.right)
      {
         putcel(newx+r.left+2,newy-STATBAR_TOTALHEIGHT+STATBAR_BAR4Y,
            dest,r,statbar_gfx,DDBLTFAST_SRCCOLORKEY);
      }
      r.top = STATBAR_BAR3Y;
      r.bottom = STATBAR_BAR3Y + STATBAR_HEIGHT; 
      r.right = time_resurrected.xcurrent;
      if (r.left < r.right)
      {
         putcel(newx+r.left+2,newy-STATBAR_TOTALHEIGHT+STATBAR_BAR3Y,
            dest,r,statbar_gfx,DDBLTFAST_SRCCOLORKEY);
      }
      r.top = STATBAR_BAR2Y;
      r.bottom = STATBAR_BAR2Y + STATBAR_HEIGHT; 
      r.right = time_selfblowing.xcurrent;
      if (r.left < r.right)
      {
         putcel(newx+r.left+2,newy-STATBAR_TOTALHEIGHT+STATBAR_BAR2Y,
            dest,r,statbar_gfx,DDBLTFAST_SRCCOLORKEY);
      }
      r.top = STATBAR_BAR1Y;
      r.bottom = STATBAR_BAR1Y + STATBAR_HEIGHT; 
      r.right = time_invisible.xcurrent;
      if (r.left < r.right)
      {
         putcel(newx+r.left+2,newy-STATBAR_TOTALHEIGHT+STATBAR_BAR1Y,
            dest,r,statbar_gfx,DDBLTFAST_SRCCOLORKEY);
      }
   }
   if (newy < PLAYFIELD_MIN_Y)
   {
      frameRect.top = frameRect.bottom - (newy - b_miny);
      y = PLAYFIELD_MIN_Y;
      clipped = true;
   }
	// je pod hornim barem
   else if (newy >= b_maxy)
   {
      clipped = true;
      frameRect.bottom = frameRect.top + PLAYFIELD_MAX_Y-newy;
      ybottom = PLAYFIELD_MAX_Y-1;
   }
	// je pod dolnim barem
   else if (clipped)
   {
      frameRect.left = frameLeft;
      frameRect.top = frameTop;
      frameRect.right = frameLeft + xsize;  
      frameRect.bottom = frameTop + ysize;  
      clipped = false;
   }

	// tak to vykresleni
   if (!clipped || frameRect.bottom > frameRect.top)  
   {
      // pasivni hrac
      if (state & NOTTAKEN)
      {
         int tt = (time&511);
         BYTE tint;
         if (tt & 256)
            tint = hardtint;
         else
            tint = litetint;
         if (gfxmask)
            puttintm(newx,y,dest,frameRect,gfxmask,tnts[tint]);
         else
            puttint(newx,y,dest,frameRect,gfx,tnts[tint]);
      }
      // neviditelny
      else if (state & INVISIBLE)
      {
         puttinted50(newx,y,dest,frameRect,gfx,(gBits == 5) ? 32767 : 65535);
      }
      // nezranitelny
      else if ((state & RESURRECTED) && (time_resurrected.timecurrent & 64))
      {
         puttinted50(newx,y,dest,frameRect,gfx,(gBits == 5) ? 32767 : 65535);
      }
      // normalni
      else
      {
         putcel(newx,y,dest,frameRect,gfx);
      }
      SET_USEDMAP(newx,y);
      SET_USEDMAP(xright,y);
      SET_USEDMAP(newx,ybottom);
      SET_USEDMAP(xright,ybottom);
   }
}

//==========================================================================
void Player::BubbleAction(int bubx, int buby)
{
   newx = bubx;
   newy = buby;
}

//==========================================================================
void Player::CheckState(int stepcnt)
{
   if (! (state & NOCREDIT))
   {
      // nezranitelnost
      if (state & RESURRECTED)
      {
         if (time_resurrected.decrease(stepcnt) == 0)
            state ^= RESURRECTED;
      }
      // neviditelnost
      if (state & INVISIBLE)
      {
         if (time_invisible.decrease(stepcnt) == 0)
            state ^= INVISIBLE;
      }
      // samopraskaci bubliny
      if (time_selfblowing.timecurrent > 0)
      {
         time_selfblowing.decrease(stepcnt);
      }
   }
	
	// automaticka aktualizace posledni viditelne pozice, pokud je viditelny
	if (!(state & (INVISIBLE | INBUBBLE)))
	{
		lastseenx = newx;
		lastseeny = newy;
	}
}

//==========================================================================
void Player::Stats(int startx)
{
   #define ATX   248  //posunuti sloupce s ciselnymi udaji
   char nums[4];// = "xx";
   int value;
   
   Font *f_normal = &qfont_y, *f_changed = &qfont_r, *f;

   // jmeno
   putString(startx,50,backbuffer,sheet.name,qfont_w);
   
   // a ted ty vlastnosti a stav
   if (state & DEAD)
   {
      putString(startx+100,210,backbuffer,"Mrtev",qfont_y);
   }
   else
   {
      if (state & NOTTAKEN)
         putString(startx+20,210,backbuffer,"Nen ve he",qfont_y);
      else
      {
         // bodovy koeficient
         value = sheet.koef;
         putString(startx,80,backbuffer,"Koeficient",qfont_y);
         numtos(value,2,nums);         
         f = (value > 1) ? f_changed : f_normal;
         putString(startx+ATX,80,backbuffer,nums,*f);     
         // nezranitelny
         if (state & RESURRECTED)
         {
            value = time_resurrected.timecurrent;
            putString(startx,120,backbuffer,"Nezraniteln",qfont_y);
            numtos(value/500,2,nums);
            f = (value > 0) ? f_changed : f_normal;
            putString(startx+ATX,120,backbuffer,nums,*f);
         }
         else
         {
            putString(startx,120,backbuffer,"Zraniteln",qfont_y);
         }
         // neviditelny
         if (state & INVISIBLE)
         {
            value = time_invisible.timecurrent;
            putString(startx,140,backbuffer,"Neviditeln",qfont_y);     
            numtos(value/500,2,nums);
            f = (value > 0) ? f_changed : f_normal;
            putString(startx+ATX,140,backbuffer,nums,*f);     
         }
         else
         {
            putString(startx,140,backbuffer,"Viditeln",qfont_y);     
         }
         // typ bublin
         if (time_selfblowing.timecurrent > 0)
         {
            putString(startx,160,backbuffer,"Kehk bubliny",qfont_y);
            value = time_selfblowing.timecurrent; 
             numtos(value/500,2,nums);
            f = (value > 0) ? f_changed : f_normal;
            putString(startx+ATX,160,backbuffer,nums,*f);
         }
         else
         {
            putString(startx,160,backbuffer,"Pevn bubliny",qfont_y);
         }
         // OBECNE VLASTNOSTI
         // rychlost (200-450)
         putString(startx,200,backbuffer,"Rychlost",qfont_y);
         value = topx/MOVE25;
         numtos(value,2,nums);
         f = (value > 8) ? f_changed : f_normal;
         putString(startx+ATX,200,backbuffer,nums,*f);
         // zrychleni (10-45)
         putString(startx,220,backbuffer,"Hbitost",qfont_y);
         value = acceleration/5+6;
         numtos(value,2,nums);
         f = (value > 10) ? f_changed : f_normal;
         putString(startx+ATX,220,backbuffer,nums,*f);
         // odraz (350-440)
         putString(startx,240,backbuffer,"Odraz",qfont_y);
         value = (jump_strength-5406)/MOVE25;
         numtos(value,2,nums);
         f = (value > 8) ? f_changed : f_normal;
         putString(startx+ATX,240,backbuffer,nums,*f);
         // rychlost bublin (500-900)
         putString(startx,260,backbuffer,"Plivn",qfont_y);
         value = bubble_speed/MOVE50;
         numtos(value,2,nums);
         f = (value > 10) ? f_changed : f_normal;
         putString(startx+ATX,260,backbuffer,nums,*f);
         // pocet bublin (6-15)
         putString(startx,280,backbuffer,"Poet bublin",qfont_y);
         value = sheet.maxenergy;
         numtos(value,2,nums);
         f = (value > 6) ? f_changed : f_normal;
         putString(startx+ATX,280,backbuffer,nums,*f);
         // repelent
         value = repelent/5;
         putString(startx,300,backbuffer,"Repelent",qfont_y);
         numtos(value,2,nums);
         f = (value > 0) ? f_changed : f_normal;
         putString(startx+ATX,300,backbuffer,nums,*f);         
      } 
   }
}

//##########################################################################
//---------------------------------- Monster -------------------------------
//##########################################################################
Monster::Monster(int onx, int ony, Property *prop) : Creature (onx,ony,prop)
{
   decision = 1 + random(1);
   type = MONSTER;
   kill_points = prop->value1;
   catch_points = prop->value2;
   bonus_cr = 0;
}

//-------------------------------------------------------------------------
// obnoveni vlastnosti pri vyadnuti z bubliny
//-------------------------------------------------------------------------
void Monster::Reincarnate()
{
   movex = 0;
   movey = 0;
   decision = decision = 1 + random(1);
   action = 0;
}

//-------------------------------------------------------------------------
// test na stabilitu
//-------------------------------------------------------------------------
int Monster::IsStable()
{
   int result,parx,pary,top_move;
   
   if (newy >= PLAYFIELD_MAX_Y) //pri testech by chmatal mimo mapu
   {
      result = AT_BOTTOM;
   }
   else
   {
      
      // podle smeru, x-bod, ktery musi byt stabilni
      if (direction == DIR_R)
         parx = newx+stable_r;
      else
         parx = newx+stable_l;
      
      pary = newy + ysize;
      top_move = topy >> SHIFT_MOVE;
      
      // je stabilni ?
      result=map_value(parx,pary,map);
      if (abs(movey)<MUL_MOVE)
      {
         result = result && !map_value(parx,pary-2,map);
         if (result && map_value(parx,pary-1,map))
         {
            newy -= 1;
            mul_y -= MUL_MOVE;
         }
      }
      else
      {
         result = result && !map_value(parx,pary-top_move-2,map);
         if (result && map_value(parx,pary-top_move-1,map))
         {
            newy -= 1;
            mul_y -= MUL_MOVE;
         }
      }
   }
   wasstable = is_stable = result;
   return result;
}

//==========================================================================
void Monster::KeyReflex()
{
   char last;
   int oldmovex;
   
   last = 0;
   if (action & SPMA_JUMP)
   {
      if (movey>0 && frames.chainid != CHAIN_FALL) 
         frames.setChain(CHAIN_FALL);
      Haste(0,gravity);
   }
   
   //rozhodovani o smeru pohybu
   //obrat pri narazu
   if (collision & (SP_COL_L|SP_COL_R))
   {
      decision^=3;
      action^=3; //tohle prihazuju navic... kvuli kralika
      collision=0;
      // je to jedno... stejne se o collision nikdo jinde nezajima
   }
   //pohyb podle rozhodnuti
   if ((decision&SCL_LEFT)==SCL_LEFT) {
      if (!action) 
         frames.setChain(CHAIN_WALK);
      action |= SPMA_LEFT;
      oldmovex = movex;
      Slow(MOVE_ACCEL,0);
      if (oldmovex >= 0 && movex < 0) 
      {
         direction = DIR_L;
         frames.setDirection(DIR_L);
      }
      last |= SCL_LEFT;
   }
   if ((decision&SCL_RIGHT)==SCL_RIGHT) {
      if (!action) 
         frames.setChain(CHAIN_WALK);
      action |= SPMA_RIGHT;
      oldmovex = movex;
      Haste (MOVE_ACCEL,0);
      if (oldmovex <= 0 && movex > 0)
      {
         direction = DIR_R;
         frames.setDirection(DIR_R);
      }
      last |= SCL_RIGHT;
   }
   if ((decision & SPMA_JUMP) && !(action & SPMA_JUMP)) {
      action |= SPMA_JUMP;
      frames.setChain(CHAIN_JUMP);
      Slow(0,JUMP_STRENGTH);
      last |= SCL_JUMP;
      decision ^= SCL_JUMP;
   }
   if (!last) {
      if (!((action & SPMA_FIRE) || (action & SPMA_JUMP)))
         frames.setChain(CHAIN_STAND);
   }
}

//==========================================================================
void Monster::BubbleAction(int bubx, int buby)
{
   newx = bubx;
   newy = buby;
}

//##########################################################################
//---------------------------------- Z o m b -------------------------------
//##########################################################################
void Zomb::KeyReflex ()
{
   char last;
   BYTE oldindex;
   
   last = 0;
   if (action & SPMA_JUMP) 
   {
      Haste(0,gravity);
   }
   
   //rozhodovani o smeru pohybu
   //a obrat pri narazu
   if (collision & 3) 
   {
      if (!(action & SPMA_JUMP))
         wannajump++;
      decision^=3;
      collision=0;
      //   action&=(!3); to je hodne divne
   }
   //pohyb podle rozhodnuti doleva
   if ((decision&SCL_LEFT)==SCL_LEFT)
   {
      Slow(MOVE_ACCEL,0);
      if (movex < 0) 
         direction = DIR_L;
      if (frames.direction == DIR_R)
      {
         oldindex = frames.index;
         frames.setDirection(DIR_L);
         frames.index = 16-oldindex;
      }
      action &= ~(SPMA_RIGHT);
      action |= SPMA_LEFT;
      last |= SCL_LEFT;
   }
   //pohyb podle rozhodnuti doprava
   if ((decision&SCL_RIGHT)==SCL_RIGHT)
   {
      Haste (MOVE_ACCEL,0);
      if (movex > 0) 
         direction = DIR_R;
      if (frames.direction == DIR_L)
      {
         oldindex = frames.index;
         frames.setDirection(DIR_R);
         frames.index = 16-oldindex;  
      }
      action &= ~(SPMA_LEFT);
      action |= SPMA_RIGHT;
      last |= SCL_RIGHT;
   }
   // poskoci!
   if (wannajump>6) {
      if (!(action & SPMA_JUMP)) {
         wannajump=0;
         action |= SPMA_JUMP;
         Slow (0,JUMP_STRENGTH);
      }
      last |= SCL_JUMP;
   }
}

//===========================================================================
void Zomb::MapReflex()
{
   IsStable();
   
   // osetreni "prochazeni zdi" doleva a doprava
   if (CantGoLR())
   {
      newx = (newx/MAP_CELX) * MAP_CELX;
      // kdyz jde doleva je treba ho posunout o jeden zpatky
      if (direction == DIR_L)
      {
         newx += MAP_CELX;
      }
      if (abs(movex) > MUL_MOVE) 
         sound_device->Play(6,newx+xsize/2);
      mul_x = (newx << SHIFT_MOVE) + MUL_HALF; //jednodussi verze
      movex = 0;
   }
   // osetreni dopadu na plosinku
   if ((action & SPMA_JUMP) && (movey > 0) && is_stable)
   {
      if (movey > 100) 
         sound_device->Play(6,newx+xsize/2);
      newy = (newy/MAP_CELY) * MAP_CELY;
      mul_y = newy << SHIFT_MOVE;
      movey = (487-movey);      //tlumeny odraz
      movey = ((movey*3)/5);
      if (movey>=-196)
      {
         movey = 0;
         action ^= SPMA_JUMP;
      }
   }
   // osetreni spadnuti na konci plosinky
   if (!(action & SPMA_JUMP) && !is_stable)
   {
      action |= SPMA_JUMP;
   }
}


//##########################################################################
//---------------------------------- Nebulus -------------------------------
//##########################################################################
void Nebulus :: KeyReflex ()
{
   char last;
   
   last = 0;
   if (action & SPMA_JUMP)
   {
      if ((movey > 0) & (frames.chainid != CHAIN_FALL)) 
         frames.setChain(CHAIN_FALL);
      Haste(0,gravity);
   }
   
   //obrat pri narazu
   if (collision&3)
   {
      decision^=3;
      collision &= (~3);
   }
   //pohyb podle rozhodnuti
   if ((decision&SCL_LEFT)==SCL_LEFT) {
      action |= SPMA_LEFT;
      Slow(MOVE_ACCEL,0);
      if (direction == DIR_R && movex < 0) 
      {
         direction = DIR_L;
         frames.setDirection(DIR_L);
      }
      last |= SCL_LEFT;
   }
   if ((decision&SCL_RIGHT)==SCL_RIGHT) {
      action |= SPMA_RIGHT;
      Haste (MOVE_ACCEL,0);
      if (direction == DIR_L && movex > 0) 
      {
         direction = DIR_R;
         frames.setDirection(DIR_R);
      }
      last |= SCL_RIGHT;
   }
   if (!(action & SPMA_JUMP)) {
      action |= SPMA_JUMP;
      frames.setChain(CHAIN_JUMP);
      //znahodneni sily odrazu
      Slow (0,2*JUMP_STRENGTH/3+random(JUMP_STRENGTH));
      last |= SCL_JUMP;
   }
}

//##########################################################################
//---------------------------------- Motylek -------------------------------
//##########################################################################
Butterfly::Butterfly(int onx, int ony, Property *prop):Monster(onx,ony,prop)
{
   direction = DIR_L;
   frames.setDirection(DIR_L);
   frames.setChain(CHAIN_STAND);
   int angle = random(360);
   movex=sin(angle);
   movey=cos(angle);
   sort_priority = 1;
}

//========================================================================
int Butterfly::Move()
{
   int result = 0;
   
   mul_x += movex;
   newx = mul_x >> SHIFT_MOVE;
   if (newx<PLAYFIELD_MIN_X) {
      newx = PLAYFIELD_MIN_X;
      mul_x = newx << SHIFT_MOVE;
      movex=-movex;
      sound_device->Play(7,newx+xsize/2);
      result |= SP_COL_L;
   }
   else if (newx>b_maxx) {
      newx = b_maxx;
      mul_x = newx << SHIFT_MOVE;
      movex=-movex;
      sound_device->Play(7,newx+xsize/2);
      result |= SP_COL_R;
   }
   mul_y += movey;
   newy = mul_y >> SHIFT_MOVE;
   if (newy<PLAYFIELD_MIN_Y) {
      newy = PLAYFIELD_MIN_Y;
      mul_y = newy << SHIFT_MOVE;
      movey=-movey;
      sound_device->Play(7,newx+xsize/2);
      result |= SP_COL_U;
   }
   else if (newy>b_maxy) {
      newy = b_maxy;
      mul_y = newy << SHIFT_MOVE;
      movey=-movey;
      sound_device->Play(7,newx+xsize/2);
      result |= SP_COL_D;
   }
   collision |= result;
   return 0;
}

//===========================================================================
void Butterfly::MapReflex()
{
}

//==========================================================================
void Butterfly::KeyReflex ()
{
}

//===========================================================================
void Butterfly::Reincarnate()
{
   int angle = random(360);
   movex = sin(angle) ;
   movey = cos(angle);
}


//##########################################################################
//---------------------------- Druhy motylek -------------------------------
//##########################################################################
void Butterfly2::MapReflex()
{
   //tady se bude resit odrazeni od plosinek
   int mapx = newx/MAP_CELX + 1;
   int mapy = (newy/MAP_CELY + 1) * MAP_XSIZE;
   BYTE res = 0;
   //zjistim, jestli se dostal na jine policko (a nebyl uz predtim ve zdi)
   // (to by melo byt ochranou proti vytuhnuti nekde ve zdi)
   if ((mapx!=oldmapx || mapy!=oldmapy) && map[oldmapx+oldmapy] == 0)
   {
      //co na nem je?
      WORD tile = map[mapx + mapy];
      WORD tilex = map[oldmapx + mapy];
      WORD tiley = map[mapx + oldmapy];
      if (tile)
      {
         //zmenilo se jen y
         if (oldmapx == mapx) 
            res = 2;
         else
            //zmenilo se jen x
            if (oldmapy == mapy) 
               res = 1;
            else
            {
               //presel krizem
               // bud jsou vedle jine plosinky
               if (tilex) 
                  res += 2;
               if (tiley) 
                  res += 1;
               // nebo je to jedina plosinka okolo
               else
                  if (!tilex) 
                     res = 3;
            }
      }
      else 
         if (tilex && tiley && oldmapx!=mapx && oldmapy!=mapy)
            res=3;
   }
   //pokud je duvod, zmenim smer pohybu a posunu, kde byl
   if (res)
   {
      sound_device->Play(7,newx+xsize/2);
      newx = oldx;
      newy = oldy;
      mul_x = mul_prevx;
      mul_y = mul_prevy;
      if (res & 1) 
         movex=-movex;
      if (res & 2) 
         movey=-movey;
   }
   else
   {
      mul_prevx = mul_x;
      mul_prevy = mul_y;
      oldmapx = mapx;
      oldmapy = mapy;
   }
}

//==========================================================================
void Butterfly2::Reincarnate()
{
   int angle = random(360);
   movex = sin(angle);
   movey = cos(angle);
   
   oldmapx = newx/MAP_CELX + 1;
   oldmapy = (newy/MAP_CELY + 1) * MAP_XSIZE;
   mul_prevx = mul_x;
   mul_prevy = mul_y;
}


//##########################################################################
//-------------------------- Horizontalni motyl ----------------------------
//##########################################################################
Butterfly3::Butterfly3(int onx, int ony, Property *prop)
:Butterfly(onx,ony,prop)
{
   if (random(1))
      movex = MOVE100;
   else 
      movex= -MOVE100;
   movey=0;
}

//==========================================================================

void Butterfly3::Reincarnate()
{
   if (random(1))
      movex = MOVE100;
   else 
      movex= -MOVE100;
   movey=0;
}

//==========================================================================
void Butterfly3::MapReflex()
{
   //tady se bude resit odrazeni od plosinek
   int mapx = newx/MAP_CELX + 1;
   int mapy = (newy/MAP_CELY + 1) * MAP_XSIZE;
   //je na novem policku do ceho narazit
   if (map[mapx + mapy])
   {
      char oldmapx=oldx/MAP_CELX + 1;
      if (!map[oldmapx + mapy])
      {
         sound_device->Play(7,newx+xsize/2);
         movex=-movex;
      }
   }
}


//#########################################################################
//--------------------------------- SymBuf --------------------------------
//#########################################################################
void SymBuf::Add(Symbol s)
{
   type <<= 8;
   type += s + 1;   
   changed = 2;
}

//=========================================================================
void SymBuf::Realize(Player *p)
{
   char * mess = NULL;
   Font f;
   Creature * m;
   Bubble * b;
   int i;
   int timeaddon;

   if (!(player1->state & NOTTAKEN) && !(player2->state & NOTTAKEN))
      timeaddon = 15000;
   else
      timeaddon = 7500;
   
   if (p->id == 1) 
      f = qfont_g;
   else
      f = qfont_b;

   //bonusy pro jednoho
   switch (type)
   {
      // docasne
   case 0x01020304 :
      p->state |= RESURRECTED;
      p->time_resurrected.increase(15000);
      mess = "Nezranitelnost";
      break;
   case 0x02020101 :
      p->state |= INVISIBLE;
      p->time_invisible.increase(15000);
      mess = "Neviditelnost";
      break;
   case 0x02020201 :
      p->time_selfblowing.increase(15000);
      mess = "Kehk bubliny";
      break;
   case 0x01010104 :
      frozen.increase(7500);
      mess = "Mrazk";
      break;
      // vlastnosti
   case 0x01010102 :
      if (p->jump_strength<MOVE650) //at nepretece
      {
         p->jump_strength += MOVE25;
         mess = "Vy vskok";
      }
      break;
   case 0x02020203 :
      if (p->acceleration<45) //at nepretece
      {
         p->acceleration += 5;
         mess = "Zven hbitosti";
      }
      break;
   case 0x03030304 :
      if (p->topx<MOVE350) //at nepretece
      {
         p->topx += MOVE25;
         p->bottomx = -p->topx;
         mess = "Vy rychlost";
      }
      break;
   case 0x04040401 :
      if (p->bubble_speed<MOVE750) //at nepretece
      {
         p->bubble_speed += MOVE50;
         mess = "Rychlej bubliny";
      }
      break;
   case 0x01020303 :
      if (p->repelent<450)
      {
         p->repelent += 50;
         mess = "Repelent";
      }
      break;
   case 0x01010202 :
      if (p->sheet.maxenergy<15) //at nepretece
      {
         p->sheet.maxenergy++;
      }
      if (p->sheet.energy<p->sheet.maxenergy)
      {
         p->sheet.energy++;
         p->sheet.energy_changed = 2;
         mess = "Bublina navc";
      }
      break;
   case 0x03030404 :
      if (p->sheet.koef<10)
      {
         p->sheet.koef ++;
         mess = "Vet vivnost";
      }
      break;
      // bodove
   case 0x01010101 :
      p->sheet.AddPoints(250);
      mess = "+250";
      break;
   case 0x02020202 :
      p->sheet.AddPoints(500);
      mess = "+500";
      break;
   case 0x03030303 :
      p->sheet.AddPoints(1000);
      mess = "+1000";
      break;
   case 0x04040404 :
      p->sheet.AddPoints(2000);
      mess = "+2000";
      break;
      // specialni
   case 0x1211 :
      mirror += 7500;
      mess = "nArUbY!";
      break;
   case 0x04030201 :
      if (player1->sheet.AddLives())
      {
         mess = "ivot navc";
      }
      break;
   case 0x01020202 :
      if (monstercnt > 0)
      {
         if (!(player1->state & NOTTAKEN) && !(player2->state & NOTTAKEN))
            timer += 15000;
         else
            timer += 7500;
         if (timer > (int)timelimit)
         {
            timelimit = timer;
            timeunit = timer / 150;
         }
         timecurrent = 0;
         time_chg = 0;
         RECT tR;
         tR.left = 600 - timer/timeunit;
         tR.right = 600;
         tR.top = 15;
         tR.bottom = 25;
         putcel(tR.left,15,backbuffer,tR,background,DDBLTFAST_NOCOLORKEY);
         putcel(tR.left,15,frontbuffer,tR,background,DDBLTFAST_NOCOLORKEY);
         mess = "30 sekund navc";
      }
      break;
   case 0x01010103 :
      for (i=0; i<monsters_ctrl.count; i++)
      {
         m = ARRAY_AT(monsters,i);
         if (m->type==MONSTER && !(ARRAY_FULL(qbubbles_ctrl)))
         {
            b = new Bubble(m->newx,m->newy,properties.member[ID_BUBBLE]);
            b->tint = p->litetint;
            b->whose = p;
            b->counter = BUB_START-1;
            if (m->newx > 300)
               b->movex = MOVE100;
            else
               b->movex = -MOVE100;
            ARRAY_ADD(qbubbles,qbubbles_ctrl,b)
         }
      }
      mess = "Lapa";
      break;
   case 0x01010301 :
      for (i=0; i<sbubbles_ctrl.count; i++)
      {
         b = ARRAY_AT(sbubbles,i);
         if (b->counter > BUB_DOMINATED)
         {
            b->counter = BUB_DOMINATED + i*50;
            b->sender = p;
         }
      }
      mess = "Praska";
      break;
   }
    
   //bonusy, ktere maji vyznam jen pro dva aktivni hrace
   if (mess==NULL)
   { //pokud uz je nejaka hlaska, nejaky bonus uz se delal, takze nic
      if (!(player1->state & NOTTAKEN) && !(player2->state & NOTTAKEN))
      {
         long swap;
         int sw;
         BYTE sc;
         switch (type)
         {
         case 0x01030103 :
            swap = player1->sheet.score;
            player1->sheet.score = player2->sheet.score;
            player2->sheet.score = swap;
            player1->sheet.score_changed = 2;
            player2->sheet.score_changed = 2;
            mess = "Vmna bod";
            break;
         case 0x01020102 :
            sw = player1->mul_x;
            player1->mul_x = player2->mul_x;
            player2->mul_x = sw;
            sw = player1->mul_y;
            player1->mul_y = player2->mul_y;
            player2->mul_y = sw;
            mess = "Vmna pozic";
            break;
         case 0x01040104 :
            // zrychleni (hbitost)
            sw = player1->acceleration;
            player1->acceleration = player2->acceleration;
            player2->acceleration = sw;
            // vyskok
            sw = player1->jump_strength;
            player1->jump_strength = player2->jump_strength;
            player2->jump_strength = sw;
            // repelent
            sw = player1->repelent;
            player1->repelent = player2->repelent;
            player2->repelent = sw;
            // rychlost bublin
            sw = player1->bubble_speed;
            player1->bubble_speed = player2->bubble_speed;
            player2->bubble_speed = sw;
            // bodovy koeficient
            sw = player1->sheet.koef;
            player1->sheet.koef = player2->sheet.koef;
            player2->sheet.koef = sw;
            // maximalni rychlost
            sw = player1->topx;
            player1->topx = player2->topx;
            player2->topx = sw;
            player1->bottomx = - (player1->topx);
            player2->bottomx = - (player2->topx);
            //
            mess = "Vmna vlastnost";
            break;
         case 0x02030203 :
            sc = player1->dik_left;
            player1->dik_left = player2->dik_left;
            player2->dik_left = sc;
            sc = player1->dik_right;
            player1->dik_right = player2->dik_right;
            player2->dik_right = sc;
            sc = player1->dik_jump;
            player1->dik_jump = player2->dik_jump;
            player2->dik_jump = sc;
            sc = player1->dik_fire;
            player1->dik_fire = player2->dik_fire;
            player2->dik_fire = sc;
            mess = "Vmna hr";
            break;
         default : return;
         }
      }
      else
      { //nepadlo to na zadny bonus...
         return;
      }
   }
   //a zarizeni vypisu hlasky
   sound_device->Play(23);
   message.Options(320,190,&f);
   message.Set(mess,&f);
   flash = 255; // 31 << 3 + zbytek
}
      
//=========================================================================
void SymBuf::Paint(Surface dest)
{
   BYTE *b;
   b = (BYTE*) &type;
   RECT r;

   r.left = x;
   r.right = r.left + 80;
   r.top = y;
   r.bottom = r.top + 20;

   // blikani po slozeni bonusu
   if (flash)
   {
     // blika!
     WORD col = (flash >> 3);
     col = col | (col<<5) | (col<<(5+gBits));
     // pozadi sym.bufferu vybarvim col
     clear(dest,col,&r);
   }
   else
   {
     // tma
     putcel(x,y,dest,r,background,DDBLTFAST_NOCOLORKEY);
   }

   int i = 0;
   while (b[i] != 0 && b[i] < 5 && i < 4)
   {
      putcel(x+(3-i)*20,y,dest,rect[b[i]-1],tiles);
      i++;
   }
}

//=========================================================================
void SymBuf::UpdateFlash()
{
  if (flash && !changed) 
  {
    flash--;
    if (flash == 0) changed = 2;
  }
}



